use vulkano::device::{Device, Queue};
use vulkano::instance::{PhysicalDevice, QueueFamily};
use vulkano::pipeline::{GraphicsPipeline, GraphicsPipelineAbstract, GraphicsPipelineBuilder};
use std::sync::Arc;
use std::ffi::CStr;
use std::path::PathBuf;
use shade_runner as sr;
use vulkano::framebuffer::{Subpass, RenderPassAbstract, Framebuffer, FramebufferAbstract};
use vulkano::pipeline::shader::{GraphicsShaderType, ShaderModule, SpecializationConstants, SpecializationMapEntry};
use vulkano::swapchain::Capabilities;
use vulkano::pipeline::vertex::SingleBufferDefinition;
use crate::util::vertex_3d::Vertex3D;
use vulkano::pipeline::depth_stencil::{DepthStencil, Stencil, StencilOp, Compare, DepthBounds};
use std::collections::HashSet;

/// Inheriting this gives private functions to grab resources
trait CompiledGraphicsPipelineResources {
    fn get_paths(filename: String, types: HashSet<ShaderType>) -> Vec<(ShaderType, PathBuf)> {
        let project_root =
            std::env::current_dir()
                .expect("failed to get root directory");

        let mut shader_path = project_root.clone();
        shader_path.push(PathBuf::from("resources/shaders/"));

        let mut paths = Vec::new();

        for shader_type in types {
            match shader_type {
                ShaderType::VERTEX => {
                    let mut shader_path = shader_path.clone();
                    shader_path.push(PathBuf::from(filename.clone() + ".vert"));
                    paths.push((shader_type, shader_path));
                }
                ShaderType::FRAGMENT => {
                    let mut shader_path = shader_path.clone();
                    shader_path.push(PathBuf::from(filename.clone() + ".frag"));
                    paths.push((shader_type, shader_path));
                }
                ShaderType::GEOMETRY => {
                    let mut shader_path = shader_path.clone();
                    shader_path.push(PathBuf::from(filename.clone() + ".geom"));
                    paths.push((shader_type, shader_path));
                }
                ShaderType::TESSELLATION => {
                    let mut shader_path = shader_path.clone();
                    shader_path.push(PathBuf::from(filename.clone() + ".tess"));
                    paths.push((shader_type, shader_path));
                }
            }
        }

        paths
    }
}

/// Typed wrapper for a u32 handle
#[derive(Clone, Debug, Default, PartialEq, Eq, Hash)]
pub struct CompiledGraphicsPipelineHandle {
    pub handle: u32
}

pub trait CompiledGraphicsPipeline {
    fn get_name(&self) -> String;
    fn get_handle(&self) -> Arc<CompiledGraphicsPipelineHandle>;
    fn get_pipeline(&self) -> Arc<dyn GraphicsPipelineAbstract + Sync + Send>;
    fn recompile(self, render_pass: Arc<dyn RenderPassAbstract + Send + Sync>)
        -> Self where Self: Sized;
}

/// Legacy ShaderType enum for single type shaders.
#[derive(PartialEq, Eq, Hash, Clone)]
pub enum ShaderType {
    VERTEX = 0,
    FRAGMENT = 1,
    GEOMETRY = 2,
    TESSELLATION_CONTROL = 3,
    TESSELLATION_EVALUATION = 4,
}


/// Typed wrapper for a u32 shader handle (index id)
#[derive(Clone, Debug, Default, PartialEq, Eq, Hash)]
pub struct CanvasShaderHandle {
    pub handle: u32
}

/// CanvasShader holds the pipeline and render pass for the input shader source
#[derive(Clone)]
pub struct GenericShader {
    graphics_pipeline: Option<Arc<dyn GraphicsPipelineAbstract + Sync + Send>>,

    handle: Arc<CompiledGraphicsPipelineHandle>,
    name: String,
    shader_types: HashSet<ShaderType>,

    device: Arc<Device>,
}

/// Gives CanvasShader the resource functions
impl CompiledGraphicsPipelineResources for GenericShader {}

/// Convenience interface so we don't have to juggle shader types
impl CompiledGraphicsPipeline for GenericShader {
    fn get_name(&self) -> String {
        self.name.clone()
    }

    fn get_handle(&self) -> Arc<CompiledGraphicsPipelineHandle> {
        self.handle.clone()
    }

    fn get_pipeline(&self) -> Arc<dyn GraphicsPipelineAbstract + Sync + Send> {
        self.graphics_pipeline.clone().unwrap()
    }

    fn recompile(self, render_pass: Arc<dyn RenderPassAbstract + Send + Sync>) -> GenericShader {
        GenericShader::new(self.name,
                           self.shader_types,
                           self.device,
                           self.handle,
                           render_pass.clone())
    }
}

impl GenericShader {

    /// This will explode when the shader does not want to compile
    pub fn new(filename: String,
               shader_types: HashSet<ShaderType>,
               device: Arc<Device>,
               handle: Arc<CompiledGraphicsPipelineHandle>,
               render_pass: Arc<dyn RenderPassAbstract + Send + Sync>) -> GenericShader {

        let filenames = GenericShader::get_paths(filename.clone(), shader_types.clone());

        // TODO: better compile message, run til successful compile




        for shader in filenames {
            match shader.0 {
                ShaderType::VERTEX => {
                    let shader = sr::load(filenames.0, filenames.1)
                        .expect("Shader didn't compile");

                    let vulkano_entry =
                        sr::parse(&shader)
                            .expect("failed to parse");
                }
                ShaderType::FRAGMENT => {

                }
                ShaderType::GEOMETRY => {

                }
                ShaderType::TESSELLATION => {

                }
            }
        }

        let fragment_shader_module: Arc<ShaderModule> = unsafe {
            let filenames1 = GenericShader::get_path(filename.clone());
            let shader1 = sr::load(filenames1.0, filenames1.1)
                .expect("Shader didn't compile");
            ShaderModule::from_words(device.clone(), &shader1.fragment.clone())
        }.unwrap();


        let vertex_shader_module: Arc<ShaderModule> = unsafe {

            let filenames1 = GenericShader::get_path(filename.clone());
            let shader1 = sr::load(filenames1.0, filenames1.1)
                .expect("Shader didn't compile");
            ShaderModule::from_words(device.clone(), &shader1.vertex.clone())
        }.unwrap();


        let frag_entry_point = unsafe {
            Some(fragment_shader_module.graphics_entry_point(CStr::from_bytes_with_nul_unchecked(b"main\0"),
                                                             vulkano_entry.frag_input,
                                                             vulkano_entry.frag_output,
                                                             vulkano_entry.frag_layout,
                                                             GraphicsShaderType::Fragment))
        };

        let vertex_entry_point = unsafe {
            Some(vertex_shader_module.graphics_entry_point(CStr::from_bytes_with_nul_unchecked(b"main\0"),
                                                           vulkano_entry.vert_input,
                                                           vulkano_entry.vert_output,
                                                           vulkano_entry.vert_layout,
                                                           GraphicsShaderType::Vertex))
        };

        let stencil = DepthStencil {
            depth_compare: Compare::Less,
            depth_write: true,
            depth_bounds_test: DepthBounds::Disabled,
            stencil_front: Stencil {
                compare: Compare::Equal,
                pass_op: StencilOp::IncrementAndWrap,
                fail_op: StencilOp::DecrementAndClamp,
                depth_fail_op: StencilOp::Keep,
                compare_mask: None,
                write_mask: None,
                reference: None,
            },
            stencil_back: Stencil {
                compare: Compare::Equal,
                pass_op: StencilOp::Invert,
                fail_op: StencilOp::Zero,
                depth_fail_op: StencilOp::Zero,
                compare_mask: None,
                write_mask: None,
                reference: None,
            },
        };


        GenericShader {
            graphics_pipeline: Some(Arc::new(GraphicsPipeline::start()

                .vertex_input(SingleBufferDefinition::<Vertex3D>::new())

                .vertex_shader(vertex_entry_point.clone().unwrap(), ShaderSpecializationConstants {
                    first_constant: 0,
                    second_constant: 0,
                    third_constant: 0.0,
                })

                .triangle_list()
                // Use a resizable viewport set to draw over the entire window
                .viewports_dynamic_scissors_irrelevant(1)

                .fragment_shader(frag_entry_point.clone().unwrap(), ShaderSpecializationConstants {
                    first_constant: 0,
                    second_constant: 0,
                    third_constant: 0.0,
                })

                .depth_stencil(stencil)

                // We have to indicate which subpass of which render pass this pipeline is going to be used
                // in. The pipeline will only be usable from this particular subpass.
                .render_pass(Subpass::from(render_pass.clone(), 0).unwrap())

                .build(device.clone())
                .unwrap())),

            device: device,
            handle: handle.clone(),
            name: filename.clone(),
            shader_types: shader_types.clone(),
        }
    }
}

#[repr(C)]
#[derive(Default, Debug, Clone)]
/// Specialization constants which can be passed to the shader. Pretty much placeholder ATM
struct ShaderSpecializationConstants {
    first_constant: i32,
    second_constant: u32,
    third_constant: f32,
}

unsafe impl SpecializationConstants for ShaderSpecializationConstants {
    fn descriptors() -> &'static [SpecializationMapEntry] {
        static DESCRIPTORS: [SpecializationMapEntry; 3] = [
            SpecializationMapEntry {
                constant_id: 0,
                offset: 0,
                size: 4,
            },
            SpecializationMapEntry {
                constant_id: 1,
                offset: 4,
                size: 4,
            },
            SpecializationMapEntry {
                constant_id: 2,
                offset: 8,
                size: 4,
            },
        ];

        &DESCRIPTORS
    }
}